import itertools

import ts_deepscan

from pathlib import Path
from typing import Iterable

from ts_python_client.commands import parse_cmd_opts_from_args

from .pm import DependencyScan


def do_scan(paths: [Path]) -> Iterable[DependencyScan]:
    """
    Imports and excutes actual scan routines
    :param paths: List of paths to be scanned
    :return: An iterable over scan results
    """
    from .pm.pypi import scan as pypi_scan
    from .pm.maven import scan as maven_scan

    for p in paths:
        p = p.resolve()

        if scan := pypi_scan(p):
            yield scan

        if scan := maven_scan(p):
            yield scan


__ds_scanner = None
__ds_dataset = None

def process_scan(scan: DependencyScan, enable_deepscan: bool, ds_args: []) -> DependencyScan:
    from ts_deepscan.cli import scan as ds_cmd

    global __ds_scanner
    global __ds_dataset

    if not __ds_scanner and enable_deepscan:
        ds_args = list(itertools.chain.from_iterable(xd.split(',') for xd in ds_args))
        ds_opts = parse_cmd_opts_from_args(ds_cmd, ds_args)

        __ds_scanner = ts_deepscan.create_scanner(**ds_opts)


    if not __ds_dataset:
        if __ds_scanner:
            for analyser in __ds_scanner.analysers:
                if __ds_dataset := getattr(analyser, 'dataset', None):
                    break
        else:
            __ds_dataset = ts_deepscan.create_dataset()

    for dep in scan.iterdeps():
        purl_v = '@' + dep.versions[0] if dep.versions else ''
        purl_ns = '/' + dep.purl_namespace if dep.purl_namespace else ''

        dep.meta['purl'] = f'pkg:{dep.purl_type}{purl_ns}/{dep.name}{purl_v}'

        if (sources := dep.files) and __ds_scanner:
            sources = [Path(src) for src in sources]
            ds_res = ts_deepscan.execute_scan(sources, __ds_scanner, title=dep.key)
            scan.deepscans[dep.key] = ds_res

        if (lic_file := dep.license_file) and lic_file.exists() and __ds_dataset:
            with lic_file.open(errors="surrogateescape") as fp:
                if lic_file_res := ts_deepscan.analyser.textutils.analyse_license_text(fp.read(), __ds_dataset):
                    dep.meta['license_file'] = lic_file_res

    return scan


# SPDX-FileCopyrightText: 2020 EACG GmbH
#
# SPDX-License-Identifier: Apache-2.0

import re
import glob
import build.util

from pathlib import Path
from typing import List, Optional, Iterable
from importlib.metadata import distribution, PackageNotFoundError

from . import DependencyScan, Dependency, License


def scan(path: Path) -> Optional[DependencyScan]:
    _scan = PypiScan(path)
    _scan.execute()

    return _scan if _scan.dependencies else None


class PypiScan(DependencyScan):
    def __init__(self, path: Path):
        super().__init__()

        self.__path = path
        self.__processed_deps = set()

        self.__dependencies = []

    @property
    def module(self) -> str:
        if len(self.__dependencies) == 1:
            return self.__dependencies[0].name
        else:
            return self.__path.name

    @property
    def moduleId(self) -> str:
        return f'pip:{self.module}'

    @property
    def dependencies(self) -> Iterable[Dependency]:
        return self.__dependencies

    def execute(self):
        stack = [self.__path]
        while len(stack) > 0:
            p = stack.pop()
            if p.is_dir():
                pkg_setup = p / "setup.py"
                if pkg_setup.exists():
                    try:
                        metadata = build.util.project_wheel_metadata(p, isolated=False)
                    except Exception as err:
                        print(f'An error occured while building packages metadata')
                        exit(2)

                    self.__dependencies.append(self._create_dep_from_metadata(metadata))
                else:
                    stack.extend([p/f for f in glob.glob('*.py', root_dir=p)])
                    stack.extend([p/d for d in glob.glob('**/', root_dir=p)])
            else:
                for pkg in _extract_imported_pkgs(p):
                    if dep := self._create_dep(pkg):
                        self.__dependencies.append(dep)


    def _create_dep(self, pkg: str) -> Optional[Dependency]:
        try:
            pkg_info = distribution(pkg)
        except  PackageNotFoundError:
            return None

        return self._create_dep_from_metadata(pkg_info.metadata)


    def _create_dep_from_metadata(self, metadata) -> Dependency:
        """
        Creates a dependency from the dist package metadate
        :param metadata: PackageMetadata
        :return:
        """
        name = metadata.get('Name', '')
        key = 'pip:' + name.lower()

        dep = Dependency(key=key, name=name, purl_type='pypi')

        if version := metadata.get('Version', None):
            dep.versions.append(version)

        if name not in self.__processed_deps:
            self.__processed_deps.add(name)

            if licence := metadata.get('License', None):
                dep.licenses.append(License(name=licence))

            dep.repoUrl = metadata.get('Download-URL', '')
            dep.description = metadata.get('Summary', '')
            dep.homepageUrl = metadata.get('Home-page', '')

            dist = distribution(name)

            #reqs = metadata.get_all('Requires-Dist', [])
            if reqs := dist.requires:
                dep.dependencies = [d for pkg in _extract_required_pkgs(reqs) if (d := self._create_dep(pkg))]

            if srcs := [Path(dist.locate_file(f)) for f in dist.files]:
                dep.files.extend(srcs)

            if lic_file := metadata.get('License-File', None):
                dep.license_file = next((f for f in dep.files if f.name == lic_file), None)

        return dep


_import_statement_regex = re.compile(r'(?:from|import) ([a-zA-Z0-9_]+).*')
def _extract_imported_pkgs(path: Path) -> List[str]:
    with path.open('r') as fp:
        try:
            data = fp.read()
            return _import_statement_regex.findall(data)
        except Exception:
            return []


_require_expr_regex = re.compile(r'([a-zA-Z0-9_\-]+).*')
def _extract_required_pkgs(reqs: List[str]) -> List[str]:
    for req in reqs:
        for pkg in _require_expr_regex.findall(req):
            yield pkg




from . import scan
from pathlib import Path

if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        s = scan(Path(sys.argv[1]))

        print([dep.files for dep in s.dependencies])
    else:
        print('No path provided')


import re
import subprocess
import typing as t

from pathlib import Path
from tempfile import TemporaryDirectory

from .. import DependencyScan, Dependency

from .pom_utils import Pom
from .tree_utils import Tree


def scan(path: Path) -> t.Optional[DependencyScan]:
    _scan = MavenScan(path)
    _scan.execute()

    return _scan if _scan.dependencies else None


class MavenScan(DependencyScan):
    def __init__(self, path: Path):
        super().__init__()

        self.__path = path
        self.__processed_deps = set()
        self.__module = None
        self.__module_id = None
        self.__tree = None

        self.__local_repo = None

        self.__dependencies = []

    @property
    def module(self) -> str:
        """returns the module name, i.e. maven artifact id"""
        return self.__module
    
    @property
    def moduleId(self) -> str:
        """returns the module id, i.e. maven key group_id:artifact_id"""
        return self.__module_id
    
    @property
    def dependencies(self) -> t.Iterable['Dependency']:
        return self.__dependencies


    def execute(self):
        self.__local_repo = self._find_local_repository()

        with TemporaryDirectory() as temp_dir:
            tree_file = Path(temp_dir)/'deps.tree'

            # Dump dependencies tree
            subprocess.run(
                ['mvn', 'dependency:tree', '-DoutputType=text', f'-DoutputFile={tree_file}'],
                cwd=self.__path,
                capture_output=True)

            # Resolve dependencies sources
            subprocess.run(
                ['mvn', 'dependency:sources'],
                cwd=self.__path,
                capture_output=True)

            if tree := Tree.from_maven_file(tree_file):
                self.__tree = tree

                dep = self._create_dep_from_node(tree)

                self.__module = dep.name
                self.__module_id = dep.key
                self.__dependencies = dep.dependencies


    def _create_dep_from_node(self, node: Tree) -> Dependency:
        # example coordinates: org.tmatesoft.svnkit:svnkit:jar:1.8.7:provided

        group_id, artifact_id, _, version, *_ = node.data.split(":")

        dep = MavenDependency(group_id=group_id, artifact_id=artifact_id, version=version, local_repo=self.__local_repo)

        if artifact_id not in self.__processed_deps:
            dep.load()

            if pkg_data := dep.package_data:
                _, _, checksum = pkg_data

                if checksum:
                    dep.checksum = checksum[1]

            if src_data := dep.sources_data:
                repo, pkg, checksum = src_data
                download_url = f'{repo}/{pkg.relative_to(self.__local_repo)}'

                sources_meta = {
                    'url': download_url
                }

                if checksum:
                    sources_meta['checksum'] = {
                        checksum[0]: checksum[1]
                    }

                dep.meta['sources'] = sources_meta

            self.__processed_deps.add(artifact_id)
            dep.dependencies = [self._create_dep_from_node(child) for child in node.children]

        return dep


    def _create_effective_pom(self) -> t.Optional[Pom]:
        with TemporaryDirectory() as temp_dir:
            pom_file = Path(temp_dir)/'effective-pom.xml'

            result = subprocess.run(
                ['mvn', 'help:effective-pom', f'-Doutput={pom_file}'],
                cwd = self.__path,
                capture_output=True
            )

            if result.returncode == 0:
                return Pom.from_maven_file(pom_file)

        return  None

    def _find_local_repository(self) -> t.Optional[Path]:
        result = subprocess.run(
            ['mvn', 'help:evaluate', '-Dexpression=settings.localRepository', '-q', '-DforceStdout=true'],
            cwd=self.__path,
            capture_output=True,
            text=True,
            encoding='utf-8'
            )

        return Path(result.stdout) if result.returncode == 0 else None


class MavenDependency(Dependency):
    def __init__(self, group_id: str, artifact_id: str, version: str, local_repo: t.Optional[Path] = None, **kwargs):
        super().__init__(key=f'mvn:{group_id}:{artifact_id}',
                         name=artifact_id,
                         versions=[version],
                         purl_type='maven',
                         purl_namespace=group_id, **kwargs)

        if local_repo:
            self.__local_repo_path = local_repo / Path(*group_id.split('.')) / Path(*artifact_id.split('.')) / version

            if self.__local_repo_path.exists() and (pom := next(self.__local_repo_path.glob('*.pom'), None)):
                self.__pom: t.Optional[Pom] = Pom.from_maven_file(pom)
            else:
                self.__pom: t.Optional[Pom] = None

    def load(self):
        if self.__pom:
            self.homepageUrl = self.__pom.url
            self.description = self.__pom.description
            self.licenses = self.__pom.licenses

        if self.__local_repo_path:
            self.files.extend(self.__local_repo_path.rglob('**'))

    @property
    def package_data(self) -> t.Optional[t.Tuple[str, Path, t.Optional[t.Tuple[str, str]]]]:
        return self._find_artifact_data('.jar')

    @property
    def sources_data(self)-> t.Optional[t.Tuple[str, Path, t.Optional[t.Tuple[str, str]]]]:
        return self._find_artifact_data('-sources.jar')


    def _find_artifact_data(self, artifact_name_suffix: str) -> t.Optional[t.Tuple[str, Path, t.Optional[t.Tuple[str, str]]]]:
        """
        Returns sources repository url, sources jar file and optionally its checksum
        :return: (sources_repo, sources_jar, (checksum_alg, checksum))
        """

        if not self.__local_repo_path:
            return None

        remote_repos = self.__local_repo_path/'_remote.repositories'
        if not remote_repos.exists():
            return None

        repos = {'central': 'https://repo.maven.apache.org/maven2'}
        if self.__pom:
            repos.update(self.__pom.repositories)

        with remote_repos.open('r') as fp:
            for line in fp:
                if m := re.search(f"([\w\d\.\-]*{artifact_name_suffix})>(.*)=", line):
                    fname = m.group(1)

                    repo = m.group(2) if m.group(2) else 'central'
                    if repo_url := repos.get(repo, None):
                        checksum = (f for f in self.__local_repo_path.glob(f'{fname}.*')
                                        if any(f.suffix.startswith(a) for a in ['.sha', '.md5']))

                        if checksum := next(checksum, None):
                            alg = checksum.suffix[1:]
                            with checksum.open('r') as checksum_fp:
                                checksum = alg, next(checksum_fp, None)

                        return repo_url, self.__local_repo_path/fname, checksum

        return None




import re
import typing as t

from pathlib import Path

class Tree:
    def __init__(self):
        self.__children = []
        self.__parent = None
        self.__data = None

    @property
    def data(self) -> t.Optional[str]:
        return self.__data
    
    @property
    def parent(self) -> t.Optional['Tree']:
        return self.__parent
    
    @property
    def children(self) -> t.List['Tree']:
        return self.__children


    @classmethod
    def from_maven_file(cls, path: Path) -> 'Tree':
        """Parses a maven-generated dependency tree file, in the format 'text'."""

        with open(path, "r") as f:
            prev_indent = 0
            prev_vertex = Tree()
            root = None

            for line in f:
                # compute indent, i.e. tree depth, by counting non-alphanumeric characters
                # at the beginning of the string, then dividing by 3
                if indent_match := re.search(r'^\W+', line):
                    new_indent = indent_match.span()[1] // 3
                else:
                    new_indent = 0

                rest = re.search(r'[a-zA-Z].*', line)

                new_vertex = Tree()
                new_vertex.__data = line[rest.start(): rest.end()].strip()

                if new_indent == 0:
                    # case 0, line has indent 0 (only happend at the start):
                    # new vertex has no parent, but is saved as a reference to the root
                    root = new_vertex

                elif new_indent > prev_indent:
                    # case 1, line is more indentet than last line:
                    # new vertex is a child of the previous vertex
                    new_vertex.__parent = prev_vertex
                    prev_vertex.__children.append(new_vertex)

                elif new_indent == prev_indent:
                    # case 2, line has the same indent as last line:
                    # new vertex is a sibling of the previous vertex
                    new_vertex.__parent = prev_vertex.__parent
                    prev_vertex.__parent.children.append(new_vertex)

                else:
                    # case 3, line has less indent than last line:
                    # new vertex is a child of the vertex n+1 levels up in the hierarchy
                    # where n is the difference in indents
                    diff = prev_indent - new_indent
                    parent = prev_vertex.__parent

                    for _ in range(diff):
                        parent = parent.parent

                    new_vertex.__parent = parent
                    parent.children.append(new_vertex)

                prev_vertex = new_vertex
                prev_indent = new_indent

        return root


import typing as t

from functools import cached_property
from pathlib import Path
from defusedxml import ElementTree

from .. import License

class Pom:
    def __init__(self):
        self.__tree: ElementTree = None
        self.__namespaces = {'xmlns': 'http://maven.apache.org/POM/4.0.0'}

    @cached_property
    def url(self) -> str:
        if node := self.__tree.find("/xmlns:url", namespaces=self.__namespaces):
            return node.text
        return ''

    @cached_property
    def description(self) -> str:
        if node := self.__tree.find("/xmlns:description", namespaces=self.__namespaces):
            return node.text
        return ''

    @cached_property
    def licenses(self) -> t.List[License]:
        names = self.__tree.findall("/xmlns:licenses/xmlns:license/xmlns:name", namespaces=self.__namespaces)
        urls = self.__tree.findall("/xmlns:licenses/xmlns:license/xmlns:url", namespaces=self.__namespaces)

        return [License(n.text.strip(), u.text.strip()) for n, u in zip(names, urls)]

    @cached_property
    def repositories(self) -> t.Dict[str, str]:
        ids = self.__tree.findall("/xmlns:repositories/xmlns:repository/xmlns:id", namespaces=self.__namespaces)
        urls = self.__tree.findall("/xmlns:repositories/xmlns:repository/xmlns:url", namespaces=self.__namespaces)
        return {n.text.strip(): u.text.strip() for n, u in zip(ids, urls)}

    @classmethod
    def from_maven_file(cls, path: Path) -> t.Optional['Pom']:
        if not path.exists():
            return None

        pom = Pom()
        pom.__tree = ElementTree.parse(path)

        return pom



import subprocess
import os
import json

from pathlib import Path, PureWindowsPath
from typing import Iterable, List, Tuple, Dict ,Optional
from enum import Enum

from defusedxml import ElementTree

from . import Dependency, DependencyScan, License

class ProjectType(Enum):
    NUSPEC = 1
    PACKAGE_REFERENCE = 2
    PACKAGES_CONFIG = 3
    SOLUTION = 4


def scan(path: Path) -> Optional[DependencyScan]:
    _scan = NugetScan(path)
    _scan.execute()

    return _scan if _scan.dependencies else None


class NugetScan(DependencyScan):
    def __init__(self, path: Path):
        super().__init__()

        self.__path = Path(path)
        self.__processed_deps = set()
        self.__module = None
        self.__module_id = None
        self.__global_packages_dir = None
        self.__nuget_command = ["nuget"]
        self.__n_fail = 0

        self.__dependencies = []


    @property
    def module(self) -> str:
        return self.__module
    
    @property
    def moduleId(self) -> str:
        return self.__module_id
    
    @property
    def dependencies(self) -> Iterable['Dependency']:
        return self.__dependencies

    
    def execute(self):
        os.chdir(self.__path)
        self.__nuget_command = self._determine_nuget_command()
        self.__global_packages_dir = self._find_global_packages_dir()

        self.__dependencies = self._process_package(self.__path)

        print(f"Scan complete, failed to process {self.__n_fail} dependenc{'y' if self.__n_fail == 1 else 'ies'}.")

    
    def _process_package(self, path: Path, depth: int = 0) -> List[Dependency]:
        print(f"Processing dependency at: {' ' * depth}{path}")

        ptype, files = self._determine_project_type(path)

        deps = []
        if ptype is ProjectType.PACKAGE_REFERENCE or ptype is ProjectType.PACKAGES_CONFIG:
            # run nuget restore with option to create a lock file
            deps = self._process_with_lock_file(files[0], depth=depth)

        elif ptype is ProjectType.NUSPEC:
            # parse nuspec file
            deps = self._create_deps_from_nuspec(files[0], depth=depth)
        
        elif ptype is ProjectType.SOLUTION:
            # extract projects from solution file, process them recursively
            deps = self._process_solution_file(files[0], depth=depth)

        return deps
    
    @staticmethod
    def _determine_project_type(path: Path) -> Tuple[ProjectType, List[Path]]:
        if files := path.glob('*.nuspec'):
            return ProjectType.NUSPEC, list(files)
        
        elif files := path.glob('*.*proj'):
            return ProjectType.PACKAGE_REFERENCE, list(files)
        
        elif files := path.glob('packages.config'):
            return ProjectType.PACKAGES_CONFIG, list(files)
        
        elif files := path.glob('*.sln'):
            return ProjectType.SOLUTION, list(files)
        else:
            raise FileNotFoundError("Could not determine project type")
        
    
    def _process_solution_file(self, solution: Path, depth: int = 0) -> List[Dependency]:
        with open(solution, "r") as f:
            projects = [line for line in f if line.strip().startswith("Project(")]

        # discard first Project line as it refers to itself
        projects = projects[1:]

        projects = [p.split("=")[-1].strip() for p in projects]
        _, paths = zip(*[p.split(",")[:2] for p in projects])

        paths = [Path(PureWindowsPath(p.strip(' "')).as_posix()) for p in paths]
        paths = [p.parent for p in paths if p.exists()]

        deps = []
        
        for path in paths:
            deps.extend(self._process_package(solution.parent / path, depth=depth))

        return deps
    

    def _process_with_lock_file(self, project_file: Path, depth: int = 0) -> List[Dependency]:
        subprocess.run(self.__nuget_command + ["restore", str(project_file), "-UseLockFile"], stdout=subprocess.PIPE)

        lockfile = project_file.parent / "packages.lock.json"

        if not lockfile.exists():
            raise FileNotFoundError("No lockfile was generated, something must have gone wrong")
        
        return self._create_deps_from_lockfile(lockfile, depth=depth)
        

    def _create_deps_from_lockfile(self, lockfile: Path, depth: int = 0) -> List[Dependency]:
        with open(lockfile, "r") as f:
            lock_dict = json.load(f)

        deps = []

        for net_target, net_target_dict in lock_dict["dependencies"].items():
            for dep_name, dep_dict in net_target_dict.items():
                if (dep_type := dep_dict["type"].lower()) in ("direct", "project"):

                    dep = Dependency(key=f"nuget:{dep_name}", name=dep_name, purl_type='nuget')

                    dep.meta[".NET target"] = net_target
                    dep.meta["dependency type"] = dep_type

                    dep_id = None
                    candidates = []

                    if dep_type == "direct":
                        dep_version = dep_dict["resolved"].lower()
                        dep.versions.append(dep_version)

                        # find package in global-packages
                        candidates = self._find_in_global_packages(dep_name, dep_version)

                        dep_id = dep.key + ":" + dep_version

                    elif dep_type == "project":
                        # dependency folder should be on the same level as project folder (i.e. sibling folder)
                        candidates = [d for d in lockfile.parent.parent.glob('*') if d.name.lower() == dep_name.lower()]
                    
                        dep_id = dep.key


                    if dep_id and dep_id not in self.__processed_deps:
                        self.__processed_deps.add(dep_id)

                        if candidates:
                            dep_dir = candidates[0]

                            dep_files = dep_dir.rglob('**')
                            dep_files = [p for p in dep_files if Path(p).is_file()]
                            dep.files.extend(dep_files)
                            
                            # recursively create dependencies of dependency
                            dep.dependencies = self._process_package(Path(dep_dir), depth=depth+1)
                        
                        else:
                            self.__n_fail += 1
                            print(f"Could not find dependency location for {dep.name}")
                            print(f"Origin: {lockfile}")

                    deps.append(dep)

        return deps
    

    def _create_deps_from_nuspec(self, nuspec: Path, depth: int = 0) -> List[Dependency]:
        ns = {"nuget": "http://schemas.microsoft.com/packaging/2013/05/nuspec.xsd"}
        tree = ElementTree.parse(nuspec)

        deps = []

        for xml_target in tree.findall("*/nuget:dependencies/nuget:group", namespaces=ns):
            target = xml_target.get("targetFramework")

            for xml_dep in xml_target.findall("nuget:dependency", namespaces=ns):
                name = xml_dep.get("id")
                version = xml_dep.get("version")

                dep_key = "nuget:" + name
                dep_id = dep_key + ":" + version

                dep = Dependency(dep_key, name)
                dep.versions.append(version)

                if not dep_id in self.__processed_deps:
                    self.__processed_deps.add(dep_id)

                    dep.meta[".NET Target"] = target
                    dep.meta["dependency type"] = "direct"

                    if candidates := self._find_in_global_packages(name, version):
                        dep_dir = candidates[0]
                        
                        dep_files = dep_dir.rglob('**')
                        dep_files = [p for p in dep_files if Path(p).is_file()]
                        dep.files.extend(dep_files)

                        if dep_nuspec := list(dep_dir.glob('*.nuspec')):
                            meta = self._metadata_from_nuspec(dep_nuspec[0])

                            dep.licenses.append(License("", meta["licenseUrl"]))
                            dep.homepageUrl = meta["projectUrl"]
                            dep.description = meta["description"]
                            dep.meta["copyright"] = meta["copyright"]

                        dep.dependencies = self._process_package(dep_dir, depth=depth+1)

                deps.append(dep)

        return deps


    def _find_global_packages_dir(self) -> Path:
        result = subprocess.run(self.__nuget_command + ["locals", "global-packages", "-list"], stdout=subprocess.PIPE)
        result = Path(result.stdout.decode("utf-8").split("global-packages: ")[1].strip())

        return result
    
    def _find_in_global_packages(self, name: str, version: str) -> List[Path]:
        """Finds all subfolders of the global-packages directory that match <name>/<version>/ (case in-sensitive)."""

        candidates = self.__global_packages_dir.glob('*/*')
        candidates = [Path(d.lower()) for d in candidates]
        candidates = [d for d in candidates if d.parts[-2] == name.lower() and d.parts[-1] == version.lower()]
    
        return candidates
    
    @staticmethod
    def _metadata_from_nuspec(nuspec: Path) -> Dict:
        ns = {"nuget": "http://schemas.microsoft.com/packaging/2013/05/nuspec.xsd"}
        tree = ElementTree.parse(nuspec)
        return_dict = {}

        for tag in ("authors", "licenseUrl", "projectUrl", "description", "copyright"):
            element = tree.find(f"*/nuget:{tag}", namespaces=ns)
            
            return_dict[tag] = element.text if element is not None else ""

        return return_dict

    @staticmethod
    def _determine_nuget_command() -> List['str']:
        try:
            command = ["nuget"]
            subprocess.run(command)

        except FileNotFoundError:
            command = ["mono", "/usr/local/bin/nuget.exe"]
            result = subprocess.run(command, stdout=subprocess.PIPE)

            if "No such file or directory" in result.stdout.decode("utf-8"):
                raise FileNotFoundError("Detected mono, but did not find nuget.exe at expected location /usr/local/bin/nuget.exe")
        
        return command
    

if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        test_scan = NugetScan(Path(sys.argv[0]))
        test_scan.execute()

        for dep in test_scan.dependencies:
            print(dep.name)
            print(dep.files)

    else:
        print('No path provided')


import os
import json
import requests
from semantic_version import Version, NpmSpec

from pathlib import Path
from typing import Iterable, Optional


from . import DependencyScan, Dependency, License


def scan(path: Path) -> Optional[DependencyScan]:
    _scan = NodeScan(path)
    _scan.execute()

    return _scan if _scan.dependencies else None


class NodeScan(DependencyScan):
    def __init__(self, path: Path):
        super().__init__()

        self.__path = Path(path)
        self.__processed_deps = set()
        self.__module = None
        self.__module_id = None
        self.__failed_requests = 0
        self.__lockfile_content = None
        self.__abs_module_path = Path(os.path.abspath(path))

        self.__lookup = {}
        self.__dependencies = []

    @property
    def module(self) -> str:
        return self.__module
    
    @property
    def moduleId(self) -> str:
        return self.__module_id
    
    @property
    def dependencies(self) -> Iterable['Dependency']:
        return self.__dependencies
    

    def execute(self):
        os.chdir(self.__path)
        os.system("npm install")

        with open(self.__path / "package-lock.json") as lockfile:
            self.__lockfile_content = json.load(lockfile)


        #self.__dependencies = self._flat_deps_from_lockfile(self.__path / "package-lock.json")

        # first build dictionary that is indexable by name
        self.__lookup = self._dict_from_lock(self.__lockfile_content)

        # then traverse graph as described in lockfile
        self.__dependencies = self._dep_from_lock(self.__lockfile_content).dependencies


    @staticmethod
    def _dict_from_lock(lock: dict) -> dict:
        deps_dict = {}

        for dep_path, dep_dict in lock["packages"].items():
            name = dep_path.split("/")[-1]
            version = dep_dict["version"]

            deps_dict.setdefault(name, {})
            deps_dict[name][version] = dep_dict
            deps_dict[name][version]["_path"] = dep_path

        return deps_dict
    

    def _dep_from_lock(self, lock: dict, package_path: str = "") -> Dependency:
        name = package_path.split("/")[-1]
        version = lock["packages"][package_path]["version"]

        dep = Dependency("npm:" + name, name)
        dep.versions.append(version)

        dep_dir = self.__abs_module_path / package_path
        dep_files = dep_dir.rglob("**")
        dep_files = [p for p in dep_files if p.is_file()]

        dep.files.extend(dep_files)

        if name + version not in self.__processed_deps:
            self.__processed_deps.add(name + version)
            print(f"Getting metadata for {name} {version}...", end="", flush=True)

            meta = self._metadata_from_registry(name, version)

            if meta:
                dep.licenses = meta.licenses
                dep.description = meta.description
                dep.homepageUrl = meta.homepageUrl
                dep.repoUrl = meta.repoUrl
                print(" Success!")
            else:
                print(" Failed!")

            for dep_name, dep_version_range in lock["packages"][package_path].get("dependencies", {}).items():
                dep_path = None

                # find appropriate dependency version
                for dep_version, dep_version_dict in self.__lookup[dep_name].items():
                    dep_path = dep_version_dict["_path"]

                    if dep_version_range == "latest":
                        break

                    try:
                        range_spec = NpmSpec(dep_version_range)

                    except ValueError:
                        range_spec = NpmSpec("")

                    if Version(dep_version) in range_spec:
                        break

                if dep_path:
                    dep.dependencies.append(
                        self._dep_from_lock(self.__lockfile_content, dep_path)
                    )

        return dep


    def _metadata_from_registry(self, name: str, version: str) -> Optional[Dependency]:
        template = "https://registry.npmjs.org/{}/{}"

        try:
            result = requests.get(template.format(name, version))

            if result.status_code != 200:
                raise requests.ConnectionError()

            meta = json.loads(result.text)
            
        except requests.ConnectionError:
            self.__failed_requests += 1
            return None


        dep = Dependency(key=f"npm:{name}", name=name, purl_type='npm')

        dep.versions.append(meta.get("version", ""))

        dep.description = meta.get("description", "")
        dep.homepageUrl = meta.get("homepage", "")
        dep.repoUrl = meta.get("repository", {}).get("url", "")

        dep.licenses.append(License(meta.get("license")))

        return dep

if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        test_scan = NodeScan(Path(sys.argv[1]))
        test_scan.execute()

        for dep in test_scan.dependencies:
            print(dep.files)
    else:
        print('No path provided')


from setuptools import setup

setup(
    name='ts-scan',
    packages=[
        'ts_scan',
        'ts_scan.pm',
        'ts_scan.pm.maven',
    ],
    version='0.2.1',
    description='TrustSource PM scanner',
    author='EACG GmbH',
    license='Apache-2.0',
    url='https://github.com/trustsource/ts-scan.git',
    download_url='',
    keywords=['scanning', 'dependencies', 'modules', 'compliance', 'TrustSource'],
    classifiers=[],
    install_requires=[
        'wheel',
        'defusedxml',
        'semanticversion',
        'build',
        'importlib-metadata',
        'alive-progress',
        'requests',
        'numpy',
        'ts-python-client==2.0.4',
        'ts-deepscan==2.0.6'
    ],
    scripts=['ts-scan'],
    entry_points={
        'console_scripts': ['ts-scan=ts_scan.cli:main'],
    }
)